// Copyright (c) 2009 Stanford University
// All rights reserved.
//
// Author: Maks Ovsjanikov, Camille Wormser
// Version: 1.0
// Description: This is a sample code for loading and displaying a pointcloud
// using the OpenGL widget in Qt. The trackball code was written by Camille Wormser.
//
// Usage: viewcloud filename
// where filename is a file containing the coordinates of all the points in the 
// x y z n1 n2 n3 format (see ../data/Model.txt for an example)
// Here n1, n2, and n3 are the components of the normal vector at the point
//
// Rotating the scene is done by holding the left mouse button,
// Zooming is done with the mouse wheel or the letters 'Z' and 'C'

#include "director.h"
#include "widget.h"

#include "cloud.h"
#include "kdtree.h"

#include <qapplication.h>
#include <qmainwindow.h>
#include <QObject>
#include <string>
#include <qmessagebox.h>

#include "iostream"

using namespace std;

string __cloud_name;

int screen_width = 500, screen_height = 500;

class MyWidget:public GLWidget
{
   Q_OBJECT
   public:
      cloud _cl;
      kdtree *_kd;
      unsigned int _stride;

   public:
  MyWidget(QWidget *parent): GLWidget(parent), _stride(3)
      {	
	 cout << "Loading Point Cloud" << endl;
	 
	 _cl.load_file(__cloud_name);

	 cout << "done, cloud size = " << _cl.size() << endl;

	 if(_cl.size()>0){
	    cout << "computing kd tree" <<  endl;
	    _kd = new kdtree(&_cl);
	    cout << "done" <<  endl;	


	    vector<unsigned int> idx;
	    vector<double> dist;
	    vector3f query(0.5,0.5,0.5);

	    int k = 10;
	    cout << "computing the " << k << " nearest neighbors of point " << query << endl;
	    _kd->find_knn(query,k,idx,dist);

	    for(unsigned int i=0;i<idx.size();i++){
	      cout << "point " << idx[i] << " is at distance " << dist[i] << endl;
	    }

	 }
	 
	 cout << "done" << endl;

	 cout << "done initialize " << endl;
      }

      ~MyWidget(){

      }

  void initializeGL (){
     GLWidget::initializeGL();
     cout << "initializing OpenGL" << endl;
     
     //set up the camera
     if(_cl.size() > 0){    
	vector3f center;
	double scale;
	_cl.get_size(scale, center);

	b.recenter(center.x,center.y,center.z);

	b.radius() *= 1./scale;
     }
  }


      void prepaintGL(){
	 GLWidget::prepaintGL();
      }  
      void mousePressEvent(QMouseEvent* e)
      {
	 GLWidget::mousePressEvent(e);

      }
  
	void draw_cloud()
      {
	 
	 glEnable(GL_POINT_SIZE);
	 glEnable(GL_POINT_SMOOTH);

	 glPointSize(1);	 
	 glColor3f(0.0,1.0,0.0);

	 glBegin(GL_POINTS);
	 for (size_t i = 0; i < _cl.size(); i += _stride)
	 {	     

	   glVertex3f(_cl[i][0],
		       _cl[i][1],
		       _cl[i][2]);
	 }
	 glEnd();
      }

      void paintGL()
      {
	 prepaintGL();

	 glColor3f(1.0, 0.0, 0.0);
         draw_cloud();

	 //postpaintGL();
      }


      void keyPressEvent (QKeyEvent* e)
      {
	 GLWidget::keyPressEvent(e);

	 if(e->key() == Qt::Key_Plus)
	    _stride++;
	 

	 if(e->key() == Qt::Key_Minus && _stride > 1)
	    _stride--;
	 

	 //	 QGLWidget::updateGL();
      }

  
};

#include "viewcloud.moc"

int main(int argc, char** argv)
{

  if(argc != 2){
    cout << "Usage: viewcloud filename" << endl;
  }
  else{
    string name(argv[1]);
    __cloud_name = name;
  }
    
   QApplication app(argc, argv);
   QMainWindow main;

   MyWidget w1(&main);
   main.setCentralWidget(&w1);
   main.resize(screen_width, screen_height);
   main.show();
   return app.exec();
}
